import multiprocessing

import click
import rich
import sentry_sdk

from decentnet.cli.keys import generate_impl
from decentnet.modules.banner.banner import orig_text
from decentnet.modules.db.base import session_scope
from decentnet.modules.db.models import OwnedKeys, AliveBeam
from decentnet.modules.migrate.migrate_agent import MigrateAgent
from decentnet.modules.monitoring.statistics import Stats
from decentnet.modules.seed_connector.SeedsAgent import SeedsAgent
from decentnet.modules.tcp.server import TCPServer

sentry_sdk.init(
    dsn="https://71d6a0d07fac5d2f072b6c7151321766@o4507850186096640.ingest.de.sentry.io/4507850892378192",
    # Set traces_sample_rate to 1.0 to capture 100%
    # of transactions for tracing.
    traces_sample_rate=1.0,
    # Set profiles_sample_rate to 1.0 to profile 100%
    # of sampled transactions.
    # We recommend adjusting this value in production.
    profiles_sample_rate=1.0,
)


@click.group()
def service():
    pass


@service.command()
@click.argument('host', type=click.STRING)
@click.argument('port', type=int)
def start(host: str, port: int):
    rich.print(orig_text)
    MigrateAgent.do_migrate()
    multiprocessing.Process(target=Stats.start_prometheus_server).start()
    rich.print("Starting DecentMesh...")
    with session_scope() as session:
        for beam in session.query(AliveBeam).all():
            session.delete(beam)
            session.commit()
        if session.query(OwnedKeys).first() is None:
            rich.print("Generating first keys for communication")
            generate_impl(private_key_file=None, public_key_file=None,
                          description="First Key", sign=True)
            generate_impl(private_key_file=None, public_key_file=None,
                          description="First Key", sign=True)
            generate_impl(private_key_file=None, public_key_file=None,
                          description="First Key", sign=True)
            generate_impl(private_key_file=None, public_key_file=None,
                          description="First Key", sign=False)

    server = TCPServer(host, port)
    server_process = multiprocessing.Process(target=server.run)
    server_process.start()
    rich.print("Connecting to DecentMesh seed nodes...")
    SeedsAgent(host, port)



import ctypes

PyLong_AsByteArray = ctypes.pythonapi._PyLong_AsByteArray
PyLong_AsByteArray.argtypes = [ctypes.py_object,
                               ctypes.POINTER(ctypes.c_uint8),
                               ctypes.c_size_t,
                               ctypes.c_int,
                               ctypes.c_int]


def int_to_bytes(num: int, buffer=None):
    byte_length = num.bit_length() // 8 + 1
    if buffer is None or len(buffer) < byte_length:
        buffer = (ctypes.c_uint8 * byte_length)()
    PyLong_AsByteArray(num, buffer, byte_length, 0, 1)
    return buffer



"""Block Byte allocation for serialization and deserialization"""
BLOCK_PREFIX_LENGTH_BYTES = 2
MAXIMUM_BLOCK_SIZE = 65536  # Bytes
INDEX_SIZE = 8
VERSION_SIZE = 1
NONCE_SIZE = 16
TIMESTAMP_SIZE = 8

MERGED_DIFFICULTY_BYTE_LEN = 26
HASH_LEN = 32



from typing import Optional, Tuple

from decentnet.modules.db.base import session_scope
from decentnet.modules.db.models import OwnedKeys


class AliasResolver:
    """
    AliasResolver is responsible for resolving the public key and key ID based on a given alias.
    """

    @classmethod
    def get_key_by_alias(cls, alias: str) -> Optional[Tuple[str, int]]:
        """
        Retrieves the public key and key ID based on the given alias.

        Args:
            alias (str): The alias to search for

        Returns:
            Optional[Tuple[str, int]]: A tuple containing the public key and key ID if found, otherwise None.
        """
        with session_scope() as session:
            key = session.query(OwnedKeys).filter_by(alias=alias).first()
            if key:
                return key.public_key, key.id
            return None



import random
import unittest

from decentnet.modules.blockchain.block import Block
from decentnet.modules.blockchain.blockchain import Blockchain
from decentnet.modules.pow.difficulty import Difficulty


class BlockchainTests(unittest.TestCase):
    def test_init(self):
        blockchain: Blockchain = Blockchain("some_msg")
        seed = blockchain.chain[0]
        print(f"Succeed {seed.compute_hash().value_as_hex()}")

    def test_insertion(self):
        blockchain: Blockchain = Blockchain("some_msg")
        seed = blockchain.chain[0]
        testdata = b"testssss"

        try:
            b = Block(1, seed.compute_hash().value, Difficulty(16, 8, 1, 8, 32), testdata)
            b.mine()
            self.assertTrue(blockchain.insert(b))
        except Exception as ex:
            self.fail(f"Thrown {ex}")
        self.assertEqual(len(blockchain.chain), 2)

    def test_insertion_long(self):
        blockchain: Blockchain = Blockchain("some_msg")

        for i in range(50):
            testdata = random.randbytes(8)
            try:
                b = Block(len(blockchain.chain),
                          blockchain.chain[-1].compute_hash().value,
                          Difficulty(16, 8, 1, random.randint(1, 8), 32), testdata)
                b.mine()
                print(f"Mined block {b} {b.ttc} ms")
                self.assertTrue(blockchain.insert(b))
            except Exception as ex:
                self.fail(f"Thrown {ex}")
        self.assertEqual(len(blockchain.chain), 51)


if __name__ == '__main__':
    unittest.main()



import logging
import os
from contextlib import contextmanager
from functools import lru_cache
from pathlib import Path

from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.consensus.local_config import DB_FILENAME, DATABASE_URL
from decentnet.modules.db.models import Base
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG)


@lru_cache()
def get_root_dir() -> Path:
    return Path(os.path.abspath(__file__)).parent.parent.parent.parent


db_file = get_root_dir() / DB_FILENAME
engine = create_engine(DATABASE_URL, pool_size=10, max_overflow=20, pool_recycle=3600)
Base.metadata.create_all(engine)


@contextmanager
def session_scope():
    """Provide a transactional scope around a series of operations."""
    session = Session(engine)
    try:
        yield session
        session.commit()
    except Exception as e:
        session.rollback()
        raise e
    finally:
        session.close()



import logging

import redis

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG)


class RedisPublisher:
    _instance = None  # Class variable to hold the single instance

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(RedisPublisher, cls).__new__(cls)
            # Initialize the instance once
            cls._instance.__init_once__(*args, **kwargs)
        return cls._instance

    def __init_once__(self, host='localhost', port=6379, db=0):
        """
        Initializes the RedisPublisher with connection to the Redis server, ensuring
        that this initialization only happens once.
        """
        try:
            # Attempt to create a Redis client and ping the server
            self.redis_client = redis.Redis(host=host, port=port, db=db)
            self.redis_client.ping()
            connection_info = f"Redis Connection Info:\nHost: {host}\nPort: {port}\nDatabase: {db}"
            logger.debug(connection_info)
        except (redis.exceptions.ConnectionError, redis.exceptions.TimeoutError) as e:
            logger.error(
                f'Redis client unable to connect, not able to relay beams. Error: {e}')

    def publish_message(self, channel: str, message: bytes):
        """
        Publishes a message to a given channel.
        """
        self.redis_client.publish(channel, message)



import datetime
import functools
import logging
import struct
from typing import Optional, Union

import rich
from prometheus_client import Gauge

from decentnet.consensus.block_sizing import MERGED_DIFFICULTY_BYTE_LEN, HASH_LEN, \
    INDEX_SIZE, VERSION_SIZE, NONCE_SIZE, TIMESTAMP_SIZE
from decentnet.consensus.blockchain_params import BlockchainParams
from decentnet.consensus.byte_conversion_constants import ENDIAN_TYPE
from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.consensus.local_config import registry
from decentnet.modules.blockchain.hasht import HashT
from decentnet.modules.compression.wrapper import CompressionWrapper
from decentnet.modules.hash_type.hash_type import MemoryHash
from decentnet.modules.logger.log import setup_logger
from decentnet.modules.pow.difficulty import Difficulty
from decentnet.modules.pow.pow import PoW
from decentnet.modules.timer.timer import Timer

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG)

data_header_ratio = Gauge('data_header_ratio', 'Header/Data Ratio', registry=registry)


class Block:
    index: int
    previous_hash: bytes = None
    version: bytes
    diff: Difficulty
    data: bytearray
    timestamp: float
    nonce: int
    _hash: Optional[MemoryHash]
    ttc: float
    signature: Optional[str] = None

    def __init__(self, index: int, prev_hash: bytes,
                 difficulty: Difficulty,
                 data: Union[bytearray, bytes]):
        self.index = index
        self.previous_hash = prev_hash
        self.data = data
        self.diff = difficulty
        self.timestamp = datetime.datetime.now().timestamp()
        self.version = BlockchainParams.block_version
        self.nonce = 0
        self.signature = None

    def __str__(self):
        result = f"Block #{self.index}\n" \
                 f"Previous Hash: {self.previous_hash.hex()[2:].zfill(HASH_LEN) if self.index != 0 else 'GENESIS BLOCK'}\n" \
                 f"Version: {int.from_bytes(self.version, ENDIAN_TYPE)}\n" \
                 f"Difficulty: {self.diff}\n" \
                 f"Data: {str(self.data)}\n" \
                 f"Timestamp: {self.timestamp}\n" \
                 f"Nonce: {self.nonce}\n"

        if hasattr(self, "_hash"):
            result += f"Hash: {self.hash.hex()[2:].zfill(HASH_LEN)}\n"

        return result

    @property
    def hash(self):
        return self.get_hash()

    def get_hash(self) -> bytes:
        barr = bytearray(self._hash.value)
        return bytes(barr)

    def compute_hash(self) -> MemoryHash:
        self._hash = MemoryHash(self.diff, self.to_bytes())
        return self._hash

    @functools.lru_cache(maxsize=256, typed=False)
    def to_bytes(self) -> bytes:
        index_bytes = self.index.to_bytes(8, byteorder=ENDIAN_TYPE, signed=False).zfill(
            INDEX_SIZE)
        version_bytes = self.version
        diff_bytes = self.diff.to_bytes().zfill(MERGED_DIFFICULTY_BYTE_LEN)
        previous_hash_bytes = self.previous_hash.zfill(HASH_LEN)
        nonce_bytes = self.nonce.to_bytes(NONCE_SIZE, byteorder=ENDIAN_TYPE,
                                          signed=False).zfill(
            NONCE_SIZE)
        timestamp_bytes = struct.pack('d', self.timestamp).zfill(TIMESTAMP_SIZE)

        raw_data = self.data
        try:
            compressed_data = CompressionWrapper.compress_lz4(
                raw_data)  # TODO: Encrypt data
        except TypeError as ex:
            logger.critical(f"Invalid data type in block {self.data}")
            raise ex

        packed_block = (index_bytes + version_bytes + diff_bytes + previous_hash_bytes +
                        nonce_bytes + timestamp_bytes + compressed_data)  # Speed up compression

        packed_block_len = len(packed_block)
        compressed_data_len = len(compressed_data)

        data_header_ratio.set(
            (packed_block_len - compressed_data_len) / compressed_data_len)

        logger.debug(f"Packed Block into {packed_block_len} bytes")
        return packed_block

    @classmethod
    def from_bytes(cls, compressed_block_bytes: bytes, compute_hash: bool = False):
        block = cls.__new__(cls)

        block_bytes = compressed_block_bytes

        cursor = 0

        block.index = int.from_bytes(block_bytes[:INDEX_SIZE], byteorder=ENDIAN_TYPE,
                                     signed=False)
        cursor += INDEX_SIZE

        block.version = block_bytes[cursor].to_bytes(VERSION_SIZE, ENDIAN_TYPE,
                                                     signed=False)
        cursor += VERSION_SIZE
        block.diff = Difficulty.from_bytes(
            block_bytes[cursor:cursor + MERGED_DIFFICULTY_BYTE_LEN])

        cursor += MERGED_DIFFICULTY_BYTE_LEN

        block.previous_hash = block_bytes[cursor:cursor + HASH_LEN]
        cursor += HASH_LEN

        block.nonce = int.from_bytes(
            block_bytes[cursor:cursor + NONCE_SIZE],
            byteorder=ENDIAN_TYPE, signed=False)
        cursor += NONCE_SIZE
        block.timestamp = cls.unpack_timestamp(
            block_bytes[cursor:cursor + TIMESTAMP_SIZE])
        cursor += TIMESTAMP_SIZE
        compressed_data = block_bytes[cursor:]

        block.data = CompressionWrapper.decompress_lz4(compressed_data)
        # Compute the hash after reconstructing the block
        if compute_hash:
            block._hash = MemoryHash(block.diff, block.to_bytes())
        return block

    @classmethod
    def unpack_timestamp(cls, timestamp_bytes: bytes) -> float:
        return struct.unpack('d', timestamp_bytes)[0]

    def mine(self, measure=True):
        logger.debug(f"Mining block #{self.index}")
        if measure:
            t = Timer()

        a = self.compute_hash()
        pow_tmp = PoW(a, self.diff)
        pow_tmp.compute()
        hash_t = HashT(a.value, pow_tmp.finished_nonce)
        if hash_t.value != pow_tmp.finished_hash.value:
            rich.print("[red]Critical! computed values not consistent[/red]")
        self.nonce = pow_tmp.finished_nonce

        if measure:
            self.ttc = t.stop()
            return hash_t, pow_tmp.finished_nonce, t.stop()
        else:
            return hash_t, pow_tmp.finished_nonce



import logging

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.modules.db.base import session_scope
from decentnet.modules.db.models import NodeInfoTable, AliveBeam
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG)


def remove_alive_beam_from_db(host: str, port: int):
    with session_scope() as session:
        node_info = session.query(NodeInfoTable).where(
            (
                    NodeInfoTable.ipv4 == host or NodeInfoTable.ipv6) == host and NodeInfoTable.port == port).first()
        if node_info:
            pub_key_base64 = node_info.pub_key
            logger.debug(f"Removing dead beam {pub_key_base64}")
            ab = session.query(AliveBeam).where(
                AliveBeam.pub_key == pub_key_base64).first()
            if ab:
                session.delete(ab)
                session.commit()


async def remove_alive_beam_from_db_w_pub_key(pub_key_base64: str):
    if not pub_key_base64:
        return
    with session_scope() as session:
        logger.debug(f"Removing dead beam {pub_key_base64}")
        ab = session.query(AliveBeam).where(AliveBeam.pub_key == pub_key_base64).first()
        if ab:
            session.delete(ab)
            session.commit()



import logging
from typing import Optional

import pyximport
import rich

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.modules.hash_type.hash_type import MemoryHash
from decentnet.modules.logger.log import setup_logger
from decentnet.modules.pow.difficulty import Difficulty

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG)

pyximport.install()
try:
    from decentnet.modules.pow.computation_fast import compute_pow  # noqa
except ImportError as ex:
    rich.print(f"[red]POW Operating in low hashrate mode, more instructions: {ex}[/red]")
    from decentnet.modules.pow.computation_slow import compute_pow


class PoW:
    def __init__(self, input_hash: MemoryHash, diff: Difficulty):
        self.finished_nonce: Optional[int] = None
        self.finished_hash: Optional[MemoryHash] = None
        self.previous_hash = input_hash
        self.difficulty = diff
        self.compute_time = None

    def compute(self, start_nonce=0):
        hash_t = self.previous_hash
        nonce = compute_pow(self.difficulty.n_bits, hash_t, start_nonce)

        self.finished_hash = hash_t
        self.finished_nonce = nonce

        return self



#cython: language_level=3
import ctypes
import sys
from math import log2

import argon2

from decentnet.consensus.blockchain_params import BlockchainParams

PyLong_AsByteArray = ctypes.pythonapi._PyLong_AsByteArray
PyLong_AsByteArray.argtypes = [ctypes.py_object,
                               ctypes.POINTER(ctypes.c_uint8),
                               ctypes.c_size_t,
                               ctypes.c_int,
                               ctypes.c_int]

def int_to_bytes(num: int, buffer=None):
    byte_length = num.bit_length() // 8 + 1
    if buffer is None or len(buffer) < byte_length:
        buffer = (ctypes.c_uint8 * byte_length)()
    PyLong_AsByteArray(num, buffer, byte_length, 0, 1)
    return bytes(buffer)

def compute_pow(n_bits, hash_t, nonce):
    _bits = len(hash_t.value) * 8 - n_bits

    value = hash_t.value_as_int()

    current_value = value

    while log2(hash_t.value_as_int()) >= _bits:
        current_value %= sys.maxsize
        current_value += 1
        nonce += 1
        hash_t.recompute(
            int_to_bytes(current_value))

    return nonce

def hash_func(data: bytes, diff):
    return argon2.hash_password_raw(data, BlockchainParams.default_salt, diff.t_cost,
                                    diff.m_cost,
                                    diff.p_cost, diff.hash_len_chars)
